from matplotlib import pyplot,gridspec,colors,patches,collections
import numpy
import warnings
import diatom.calculate as calculate
import diatom.hamiltonian as hamiltonian
from scipy import constants

h = constants.h

def make_segments(x, y):
    ''' segment x and y points

    Create list of line segments from x and y coordinates, in the correct format for LineCollection:
    an array of the form   numlines x (points per line) x 2 (x and y) array

    Args:
        x,y (numpy.ndarray -like ) - points on lines

    Returns:
        segments (numpy.ndarray) - array of numlines by points per line by 2

    '''

    points = numpy.array([x, y]).T.reshape(-1, 1, 2)
    segments = numpy.concatenate([points[:-1], points[1:]], axis=1)

    return segments

def colorline(x, y, z=None, cmap=pyplot.get_cmap('copper'),
                norm=pyplot.Normalize(0.0, 1.0), linewidth=3, alpha=1.0,
                legend=False,ax=None):
    '''Plot a line shaded by an extra value.


    Plot a colored line with coordinates x and y
    Optionally specify colors in the array z
    Optionally specify a colormap, a norm function and a line width

    Args:
        x,y (list-like): x and y coordinates to plot

    kwargs:
        z (list): Optional third parameter to colour lines by
        cmap (matplotlib.cmap): colour mapping for z
        norm (): Normalisation function for mapping z values to colours
        linewidth (float): width of plotted lines (default =3)
        alpha (float): value of alpha channel (default = 1)
        legend (Bool): display a legend (default = False)
        ax (matplotlib.pyplot.axes): axis object to plot on

    Returns:
        lc (Collection) - collection of lines
        
    '''
    if ax == None:
        ax = pyplot.gca()

    # Default colors equally spaced on [0,1]:
    if z is None:
        z = numpy.linspace(0.0, 1.0, len(x))

    # Special case if a single number:
    if not hasattr(z, "__iter__"):  # to check for numerical input -- this is a hack
        z = numpy.array([z])

    z = numpy.asarray(z)

    segments = make_segments(x, y)
    lc = collections.LineCollection(segments, array=z, cmap=cmap, norm=norm,#change
                                    linewidth=linewidth,zorder=1.25)

    ax.add_collection(lc)

    return lc

def export_energy(fname,energy,fields=None,labels=None,
                                headers=None,dp=6,format=None):
    ''' Export Energies in spreadsheet format.

    This exports the energy of the states for a calculation in a human-readable spreadsheet format.

    Currently only saves .csv files.

    Args:
        fname (string) - file name to save, appends .csv if not present.
        Energy (numpy.ndarray) - Energies to save

    Kwargs:
        Fields (numpy.ndarray) - Field variables used in calculation
        labels (numpy.ndarray) - labels for states
        headers (list of strings) - header for each of the labels in labels
        dp (float) - number of decimal places to use for output (default =6)
        format (list of strings) - list of formats passed to numpy.savetxt for labels
    '''
    # some input sanitisation, ensures that the fname includes an extension
    if fname[-4:]!=".csv":
        fname = fname+".csv"
    dp = int(numpy.round(dp))

    # check whether the user has given labels and headers or  just one
    lflag = False
    hflag = False

    if labels != None:
        labels = numpy.array(labels)
        lflag = True
    else:
        labels=[]

    if headers != None:
        hflag = True
    else:
        headers = []

    # all this is just checking whether there are headers and labels, just
    #labels, just headers or neither.

    if not hflag and lflag:
        warnings.warn("using default headers for labels",UserWarning)
        headers = ["Label {:.0f}".format(x) for x in range(len(labels[0,:]))]

    elif hflag and not lflag:
        warnings.warn("headers given without labels",UserWarning)
        headers =[]

    if len(headers) != labels.shape[0]:
        warnings.warn("Not enough headers given for chosen labels",UserWarning)
        headers = ["Label {:.0f}".format(x) for x in range(len(labels[:,0]))]

    # Now to write a string to make the output look nice. For simplicity we say
    # that all the  labels must be given to 1 dp
    if format==None:
        format = ["%.1f" for x in range(len(labels[:,0]))]

    # now just make the one for the main body of the output file. Specified by
    # the dp argument.
    if len(energy.shape)>1:
        format2 = ["%."+str(dp)+"f" for x in range(len(energy[:,0]))]
    else:
        format2 = ["%."+str(dp)+"f"]
    #numpy needs only one format argument
    format.extend(format2)

    headers =','.join(headers)


    headers = ','.join(["Labels" for l in range(labels.shape[0])])+",Energy (Hz)\n"+headers

    if type(fields) != type(None):
        energy = numpy.insert(energy,0,fields.real,axis=1)
        labels = numpy.insert(labels,0,[-1 for x in range(labels.shape[0])],axis=1)

    output = numpy.row_stack((labels,energy))
    numpy.savetxt(fname,output.T,delimiter=',',header = headers,fmt=format)

def export_state_comp(fname,Nmax,I1,I2,states,labels=None,
                                headers=None,dp=6,format=None):
    ''' function to export state composition in a human-readable format
    along the first row are optional headers and the labels for the basis States
    in the uncoupled basis.

    the user can supply optional labels for the states in a (x,y) list or array
    where y is the number of states and x is the number of unique labels, for
    instance a list of the N quantum  number for each state.

    they can also (optionally) supply a (x,1) list to include custom headers
    in the first row. If the labels kwarg is included and headers is not,
    then non-descriptive labels are used to ensure correct output.

    by default the output is given to 6 decimal places (truncated) this can be
    adjusted using the kwarg dp

    Args:
        fname (string) : the filename and path to save the output file
        Nmax (int/float) : the maximum value of N used in the calculation
        I1,I2 (float) : the nuclear spin quantum numbers of nucleus 1 and 2
        States (N,M) ndarray : eigenstates stored in an (N,M) ndarray, N is the
                                number of eigenstates. M is the number of basis
                                states.
    kwargs:
        labels (N,X) ndarray : ndarray containing X labels for each of the N states
        headers (X) ndarray-like : Ndarray-like containing descriptions of the labels
        dp (int) : number of decimal places to output the file to [default = 6]
        format (list) :  list of strings for formatting the headers. Defaults to 1 dp.

    '''

    # some input sanitisation, ensures that the fname includes an extension
    if fname[-4:]!=".csv":
        fname = fname+".csv"
    dp = int(numpy.round(dp))

    # check whether the user has given labels and headers or  just one
    lflag = False
    hflag = False

    if labels != None:
        labels = numpy.array(labels)
        lflag = True

    if headers != None:
        hflag = True
    else:
        headers = []


    # create labels for basis states from Generate_vecs
    # first step is to recreate the angular momentum operators
    N, I1,I2 = hamiltonian.generate_vecs(Nmax,I1,I2)

    # each basis state is an eigenstate of N^2, so N^2 is diagonal in our basis
    # with eigenvalues N(N+1)

    N2 = numpy.round([calculate.solve_quadratic(1,1,-1*x) for x in numpy.diag(hamiltonian.vector_dot(N,
                                                                N))],0).real

    # they are also eigenstates of Nz, I1z and I2z which are diagonal
    # in the basis that we constructed.

    MN = numpy.round(numpy.diag(N[2]),0).real
    M1 = numpy.round(numpy.diag(I1[2]),1).real
    M2 = numpy.round(numpy.diag(I2[2]),1).real

    # Now we create a list of each of the values in the right place
    state_list = ["({:.0f} : {:.0f} : {:.1f} : {:.1f})".format(N2[i],
                                    MN[i],M1[i],M2[i]) for i in range(len(MN))]
    # all this is just checking whether there are headers and labels, just
    #labels, just headers or neither.

    if not hflag and lflag:
        warnings.warn("using default headers for labels",UserWarning)
        headers = ["Label {:.0f}".format(x) for x in range(len(labels[0,:]))]

    elif hflag and not lflag:
        warnings.warn("headers given without labels",UserWarning)
        headers =[]

    if len(headers) != labels.shape[0]:
        warnings.warn("Not enough headers given for chosen labels",UserWarning)
        headers = ["Label {:.0f}".format(x) for x in range(len(labels[:,0]))]

    # Now to write a string to make the output look nice. For simplicity we say
    # that all the  labels must be given to 1 dp
    if format==None:
        format = ["%.1f" for x in range(len(headers))]

    # now just make the one for the main body of the output file. Specified by
    # the dp argument.
    format2 = ["%."+str(dp)+"f" for x in range(len(state_list))]

    #numpy needs only one format argument
    format.extend(format2)

    headers.extend(state_list)
    headers =','.join(headers)
    headers = ','.join(["Labels" for l in range(labels.shape[0])])+",States in (N:MN:M1:M2) basis\n"+headers
    states=numpy.transpose(states)#changde	
    output = numpy.insert(states.real,0,labels.real,axis=0)
    numpy.savetxt(fname,output.T,delimiter=',',header = headers,fmt=format)


def transition_plot(energies,states,gs,Nmax,I1,I2,TDMs=None,
            pm = +1, Offset=0,fig=pyplot.gcf(),
            log=False,minf=None,maxf=None,prefactor=1e-3,col=None):

    ''' Create a TDM plot

    this function plots a series of energy levels and their transition dipole
    moments from a given ground state. In this version a lot of the plotting style
    is fixed.

    Args:
        energies (numpy.ndarray) - array of energy levels
        states (numpy.ndarray) - array of states corresponding to energies such that E[i] -> States[:,i]
        gs (int) - index for ground state of interest
        Nmax (int) - maximum rotational quantum number to include
        I1, I2 (float) - nuclear spins of nuclei 1 and 2
    Kwargs:
        TDMs (list of numpy.ndarray) - optional precomputed transition dipole moments in [sigma-,pi,sigma+] order
        pm (float) - flag for if the transition increases or decreases N (default = 1)
        Offset (float) - yaxis offset (default = 0)
        fig (matplotlib.pyplot.figure) - figure object to draw on
        log (bool) - use logarithmic scaling for TDM plots
        minf (float) - minimum frequency to show
        maxf (float) - maximum frequency to show
        prefactor (float) - scaling factor for all energies
        col (list) - list of colours for lines (must be at least length 3 )

    '''

    gray ='xkcd:lightgold'
    gray='#fddc5c'
    if col == None:
        cc=['#61e160','#fe2f4a','#155084']
        green=cc[0]
        red=cc[1]
        blue=cc[2]
        #green ='xkcd:darkgreen'
        #red ='xkcd:maroon'
        #blue ='xkcd:azure'

        col=[red,blue,green]

    if TDMs == None and (Nmax == None or I1 == None or  I2 == None):
        raise RuntimeError("TDMs  or Quantum numbers must be supplied")

    elif (Nmax == None or I1 == None or  I2 == None):
        TDMs = numpy.array(TDMs)
        dm = TDMs[0,:]
        dz = TDMs[1,:]
        dp = TDMs[2,:]
    elif TDMs == None:
        dm = numpy.round(calculate.transition_dipole_moment(Nmax,I1,I2,+1,states,gs),6)
        dz = numpy.round(calculate.transition_dipole_moment(Nmax,I1,I2,0,states,gs),6)
        dp = numpy.round(calculate.transition_dipole_moment(Nmax,I1,I2,-1,states,gs),6)

    if abs(pm)>1:
        pm = int(pm/abs(pm))

    widths = numpy.zeros(4)+1
    widths[-1] = 1.4

    fig.set_figheight(8)
    fig.set_figwidth(6)

    grid= gridspec.GridSpec(2,4,width_ratios=widths)

    N,MN = calculate.label_states_N_MN(states,Nmax,I1,I2)
    #find the ground state that the user has put in

    N0 = N[gs]

    energies = energies-energies[gs]
    lim =10

    l1 = numpy.where(N==N0)[0]
    l2 = numpy.where(N==N0+pm)[0]


    if minf == None:

        emin = numpy.amin(energies[l2])
        minf = prefactor*(emin)/h - Offset

    if maxf == None:

        emax = numpy.amax(energies[l2])
        maxf = prefactor*(emax)/h - Offset

    if pm == 1:
        ax0 = fig.add_subplot(grid[1,:-1])
        ax = []
        for j in range(3):
            if j ==0:
                ax.append(fig.add_subplot(grid[0,j],zorder=1))
            else:
                ax.append(fig.add_subplot(grid[0,j],sharey=ax[0],zorder=1))

    elif pm == -1:
        ax0 = fig.add_subplot(grid[0,:-1])
        ax = []
        for j in range(3):
            if j ==0:
                ax.append(fig.add_subplot(grid[1,j],zorder=1))
            else:
                ax.append(fig.add_subplot(grid[1,j],sharey=ax[0],zorder=1))


    #plotting the energy levels for ground state
    for l in l1:
        f =prefactor*(energies[l])/h 
        if l ==gs:
            ax0.plot([-lim,lim],[f,f],color='k',zorder=1.2)
        else:
            ax0.plot([-lim,lim],[f,f],color=gray,zorder=0.8)
    lbl = ['$\sigma_-$',"$\pi$","$\sigma_+$"]

    for j,axis in enumerate(ax):        
    #plotting for excited state
        for l in l2:
            f = prefactor*(energies[l])/h - Offset
            if dz[l]!=0 and j==1:
                axis.plot([-lim,lim],[f,f],color=blue,zorder=1.2)
            elif dp[l] !=0 and j ==2:
                axis.plot([-lim,lim],[f,f],color=green,zorder=1.2)
            elif dm[l] !=0 and j ==0:
                axis.plot([-lim,lim],[f,f],color=red,zorder=1.2)
            else:
                axis.plot([-lim,lim],[f,f],color=gray,zorder=0.8)
        if j ==0 :
            axis.tick_params(labelbottom=False,bottom=False,which='both')
        else:
            axis.tick_params(labelleft=False,left=False,labelbottom=False,
                        bottom=False,which='both')
        axis.set_xlim(-lim,lim)
        axis.set_title(lbl[j],color=col[j])

    # set the ticks so that only the left most has a frequency/energy axis
    # and none have an x axis

    ax0.tick_params(labelbottom=False,bottom=False,which='both')
    ax0.set_xlim(-lim,lim)

    #add the bar plot axis
    ax_bar = fig.add_subplot(grid[0,-1],sharey = ax[0])
    ax_bar.tick_params(labelleft=False,left=False, which='both')

    #fix the ROI to be 200 kHz around the state the user has chosen
    if gs == 0:
        ax0.set_ylim(-20e3*prefactor, 180e3*prefactor)
    else:
        ax0.set_ylim(-20e3*prefactor, 180e3*prefactor)

    #normalise function, returns a number between 0 and 1
    Norm = colors.LogNorm(vmin=1e-3,vmax=1,clip=True)
    #how thick should a line be?
    max_width = 2

    #setting where and how far apart the lines should all be in data coords

    ax1 = ax[0]
    ax2 = ax[1]
    ax3 = ax[2]

    disp = ax2.transData.transform((-lim,0))
    x1a = ax0.transData.inverted().transform(disp)[0]

    disp = ax2.transData.transform((lim,0))
    x1b = ax0.transData.inverted().transform(disp)[0]

    Nz = len(numpy.where(dz[l2]!=0)[0])
    iz = 0

    deltax = (x1b-x1a)/(Nz+1)
    x0 = x1a+deltax

    disp = ax3.transData.transform((-lim,0))
    y1a = ax0.transData.inverted().transform(disp)[0]

    disp = ax3.transData.transform((lim,0))
    y1b = ax0.transData.inverted().transform(disp)[0]

    Np = len(numpy.where(dp[l2]!=0)[0])
    ip =0

    deltay = (y1b-y1a)/(Np+1)
    y0 = y1a+deltay

    disp = ax1.transData.transform((-lim,0))
    z1a = ax0.transData.inverted().transform(disp)[0]

    disp = ax1.transData.transform((lim,0))
    z1b = ax0.transData.inverted().transform(disp)[0]

    Nm = len(numpy.where(dm[l2]!=0)[0])
    im = 0

    deltaz = (z1b-z1a)/(Nm+1)
    z0 = z1a+deltaz
    f = prefactor*(energies)/h-Offset
    
    for j,d in enumerate(dz):
        #this block of code plots the dipole moments (or transition strengths
        if j<max(l2) and f[j]<maxf:
            if abs(d)>0 and j<max(l2):
                width = max_width*Norm(3*numpy.abs(d)**2)
                x = x0 +iz*deltax
                # makes sure that the line is perfectly vertical in display coords
                disp = ax0.transData.transform((x,0))
                x2 = ax2.transData.inverted().transform(disp)[0]
    
                p = patches.ConnectionPatch((x,0),(x2,f[j]),coordsA='data',coordsB='data',
                                                axesA=ax0,axesB=ax2,zorder=5,color='k',
                                                lw=width) #line object
                ax2.add_artist(p) # add line to axes
                iz+=1
                #bar plot for transition strengths. Relative to spin-stretched TDM
                ax_bar.barh(f[j],numpy.abs(d),color=blue,height=1e4*prefactor)
    
            d=dp[j]
            if abs(d)>0 and j<max(l2):
                width = max_width*Norm(3*numpy.abs(d)**2)
                y= y0 +ip*deltay
                # makes sure that the line is perfectly vertical in display coords
    
                disp = ax0.transData.transform((y,0))
                y2 = ax3.transData.inverted().transform(disp)[0]
    
                p = patches.ConnectionPatch((y,0),(y2,f[j]),coordsA='data',coordsB='data',
                                                axesA=ax0,axesB=ax3,zorder=5,color='k',
                                                lw=width) #line object
                ax3.add_artist(p)
                ip+=1
                #bar plot for transition strengths. Relative to spin-stretched TDM
                ax_bar.barh(f[j],numpy.abs(d),color=green,height=1e4*prefactor)
    
            d=dm[j]
            if abs(d)>0 and j<max(l2):
                width = max_width*Norm(3*numpy.abs(d)**2)
                z = z0 +im*deltaz
                # makes sure that the line is perfectly vertical in display coords
    
                disp = ax0.transData.transform((z,0))
                z2 = ax1.transData.inverted().transform(disp)[0]
    
                p = patches.ConnectionPatch((z,0),(z2,f[j]),coordsA='data',coordsB='data',
                                                axesA=ax0,axesB=ax1,zorder=5,color='k',
                                                lw=width)#line object
                ax1.add_artist(p)
                im +=1
                #bar plot for transition strengths. Relative to spin-stretched TDM
                ax_bar.barh(f[j],numpy.abs(d),color=red,height = 1e4*prefactor)

    #setup log axes for axis 4 (bar plots)
    if log:
        ax_bar.set_xscale('log')

        ax_bar.set_xticks([1e-6,1e-3,1])
        ax_bar.set_xticks([1e-5,1e-4,1e-2,1e-1],minor=True)

        ax_bar.set_xticklabels(["10$^{-6}$","10$^{-3}$","1"])
        ax_bar.set_xticklabels(["","","",""],minor=True)

    # now to rescale the other axes so that they have the same y scale
    ax1.set_ylim(minf-20e3*prefactor,maxf+20e3*prefactor)
    grid.set_height_ratios([(maxf-minf)+40e3*prefactor,200e3*prefactor])
    pyplot.subplots_adjust(hspace=0.1)
    grid.update()

    #add some axis labels
    ax0.set_ylabel("Energy/$h$ (kHz)")

    if Offset != 0:
        ax[0].set_ylabel("Energy/$h$ (kHz) - {:.3f} MHz".format(Offset+minf-20*1e3*prefactor))
    else:
        ax[0].set_ylabel("Energy/$h$ (Hz)")

    ax_bar.set_xlabel("TDM ($d_0$)")




import diatom.hamiltonian as hamiltonian
import numpy
import scipy.constants
from sympy.physics.wigner import wigner_3j

###############################################################################
# Start by definining a bunch of constants that are needed for the code       #
###############################################################################

'''
    Important note!

    All units in this code are SI i.e. elements in the Hamiltonian have units
    of Joules. Outputs will be on the order of 1e-30

'''

h = scipy.constants.h
muN = scipy.constants.physical_constants['nuclear magneton'][0]
bohr = scipy.constants.physical_constants['Bohr radius'][0]
eps0 = scipy.constants.epsilon_0
c = scipy.constants.c


def solve_quadratic(a,b,c):
    ''' Solve a quadratic equation

    for a*x^2+b*x+c=0 this is a simple function to solve the quadratic formula for x. returns the most
    positive value of x supported.

    Args:
        a,b,c (floats) - coefficients in quadratic

    Returns:
        x (float) - maximum value of x supported by equation

    '''
    x1 = (-b+numpy.sqrt((b**2)-(4*(a*c))))/(2*a)
    x2 = (-b-numpy.sqrt((b**2)-(4*(a*c))))/(2*a)
    return max([x1,x2])

def label_states_N_MN(states,Nmax,I1,I2,locs=None):
    ''' Label states by N,MN

    This function returns two lists: the input states labelled by N and MN
    in the order that they are provided. The returned numbers will only be good
    if the state is well -represented in the decoupled basis.

    Optionally can return the quantum  numbers for a subset if the locs kwarg
    is provided. Each element in the list locs corresponds to the index for the
    states to label.

    Args:

        States (Numpy.ndarray) - array of eigenstates, from linalg.eig
        Nmax (int) - maximum rotational state in calculation
        I1 , I2 (float) - nuclear spin quantum numbers

    kwargs:
        locs (list of ints) - list of indices of states to label

    Returns:
        Nlabels,MNlabels (list of ints) - list of values of N,MN

    '''
    if locs != None:
        states = states[:,locs]
    N, I1,I2 = hamiltonian.generate_vecs(Nmax,I1,I2)#change

    N2 = hamiltonian.vector_dot(N,N)#change
    Nz = N[2]

    Nlabels = numpy.einsum('ik,ij,jk->k',numpy.conj(states),N2,states)
    Nlabels = numpy.round([solve_quadratic(1,1,-1*x)  for x in Nlabels],0).real

    MNlabels = numpy.round(numpy.einsum('ik,ij,jk->k',
                                    numpy.conj(states),Nz,states),0).real

    return Nlabels,MNlabels

def label_states_I_MI(states,Nmax,I1,I2,locs = None):
    ''' Label states by I,MI

    This function returns two lists: the input states labelled by I and MI
    in the order that they are provided. The returned numbers will only be good
    if the state is well -represented in the decoupled basis.

    Optionally can return the quantum  numbers for a subset if the locs kwarg
    is provided. Each element in the list locs corresponds to the index for the
    states to label.

    Args:
        States (Numpy.ndarray) - array of eigenstates, from linalg.eig
        Nmax (int) - maximum rotational state in calculation
        I1 , I2 (float) - nuclear spin quantum numbers

    kwargs:
        locs (list of ints) - list of indices of states to label

    Returns:
        Ilabels,MIlabels (list of ints) - list of values of I,MI

    '''
    if locs != None:
        states = states[:,locs]

    #I = I1 + I2 change

    N, I1,I2 = hamiltonian.generate_vecs(Nmax,I1,I2)#change
    
    I = I1 + I2#change

    I2 = hamiltonian.vector_dot(I,I)#change

    Iz = I[2]

    Ilabels = numpy.einsum('ik,ij,jk->k',numpy.conj(states),I2,states)
    Ilabels = numpy.round([solve_quadratic(1,1,-1*x)  for x in Ilabels],1).real

    MIlabels = numpy.round(numpy.einsum('ik,ij,jk->k',
                                    numpy.conj(states),Iz,states),1).real

    return Ilabels,MIlabels

def label_states_F_MF(states,Nmax,I1,I2,locs=None):
    ''' Label states by F,MF

    This function returns two lists: the input states labelled by F and MF
    in the order that they are provided. The returned numbers will only be good
    if the state is well -represented in the decoupled basis.

    Optionally can return the quantum  numbers for a subset if the locs kwarg
    is provided. Each element in the list locs corresponds to the index for the
    states to label.

    Args:
        States (Numpy.ndarray) - array of eigenstates, from linalg.eig
        Nmax (int) - maximum rotational state in calculation
        I1 , I2 (float) - nuclear spin quantum numbers

    kwargs:
        locs (list of ints) - list of indices of states to label

    Returns:
        Flabels,MFlabels (list of ints) - list of values of F,MF

    '''

    if locs != None:
        states = states[:,locs]

    N, I1,I2 = hamiltonian.generate_vecs(Nmax,I1,I2)

    F = N + I1 + I2

    F2 = hamiltonian.vector_dot(F,F)

    Fz = F[2]

    Flabels = numpy.einsum('ik,ij,jk->k',numpy.conj(states),F2,states)
    Flabels = numpy.round([solve_quadratic(1,1,-1*x)  for x in Flabels],1).real

    MFlabels = numpy.round(numpy.einsum('ik,ij,jk->k',
                                    numpy.conj(states),Fz,states),1).real

    return Flabels,MFlabels

def dipole(Nmax,I1,I2,d,M):
    ''' Generates the induced dipole moment operator for a Rigid rotor.
    Expanded to cover state  vectors in the uncoupled hyperfine basis.

    Args:
        Nmax (int) - maximum rotational states
        I1,I2 (float) - nuclear spin quantum numbers
        d (float) - permanent dipole moment
        M (float) - index indicating the helicity of the dipole field

    Returns:
        Dmat (numpy.ndarray) - dipole matrix
    '''
    shape = numpy.sum(numpy.array([2*x+1 for x in range(0,int(Nmax+1))]))
    dmat = numpy.zeros((shape,shape),dtype= numpy.complex)
    i =0
    j =0
    for N1 in range(0,int(Nmax+1)):
        for M1 in range(N1,-(N1+1),-1):
            for N2 in range(0,int(Nmax+1)):
                for M2 in range(N2,-(N2+1),-1):
                    dmat[i,j]=d*numpy.sqrt((2*N1+1)*(2*N2+1))*(-1)**(M1)*\
                    wigner_3j(N1,1,N2,-M1,M,M2)*wigner_3j(N1,1,N2,0,0,0)
                    j+=1
            j=0
            i+=1

    shape1 = int(2*I1+1)

    shape2 = int(2*I2+1)

    dmat = numpy.kron(dmat,numpy.kron(numpy.identity(shape1),
                                                    numpy.identity(shape2)))

    return dmat

def transition_dipole_moment(Nmax,I1,I2,M,states,gs,locs=None):
    ''' calculate TDM between gs and States

    Function to calculate the Transition Dipole Moment between a state  gs
    and a range of states. Returns the TDM in units of the permanent dipole
    moment (d0).

    Args:
        Nmax (int): Maximum rotational quantum number in original calculations
        I1,I2 (float): nuclear spin quantum numbers
        M (float): Helicity of Transition, -1 = S+, 0 = Pi, +1 = S-
        States (numpy.ndarray): matrix for eigenstates of problem output from numpy.linalg.eig
        gs (int): index of ground state.

    kwargs:
        locs (list of ints): optional argument to calculate for subset of States, should be an
                array-like.

    Outputs:
        TDM(list of floats) - transition dipole moment between gs and States
    
    '''

    dipole_op = dipole(Nmax,I1,I2,1,M)
    if type(gs) == tuple:#change
        '''
        gs is usually produced by np.where() or np.nonzero(). And the outputs of those functions are
        tuples. This if condition ensures gs is an int.
        If there are more than one index in gs, only take the first one
        '''
        gs=gs[0][0]#change

    gs = numpy.conj(states[:,gs])
    if locs != None :
        states =  states[:,locs]

    tdm =  numpy.einsum('i,ij,jk->k',gs,dipole_op,states).real

    return tdm


def magnetic_moment(States, Nmax, Consts):
    '''Returns the magnetic moments of each eigenstate
    
    Args:
        States (numpy.ndarray): matrix for eigenstates of problem output from numpy.linalg.eig
        Nmax (int): Maximum rotational quantum number in original calculations
        Consts: Dictionary of constants for the molecular to be calculated
        
    '''
    
    muz = -1*hamiltonian.zeeman_ham(Nmax,Consts['I1'],Consts['I2'],Consts)
    
    mu =numpy.einsum('ijk,jl,ilk->ik',
            numpy.conjugate(States),muz,
            States)
    return mu


def electric_moment(States, Nmax, Consts):
    '''Returns the electric dipole moments of each eigenstate
    
    Args:
        States (numpy.ndarray): matrix for eigenstates of problem output from numpy.linalg.eig
        Nmax (int): Maximum rotational quantum number in original calculations
        Consts: Dictionary of constants for the molecular to be calculated
        
    '''
    
    dz = -1*hamiltonian.dc(Nmax,Consts['d0'],Consts['I1'],Consts['I2'])
    
    d =numpy.einsum('ijk,jl,ilk->ik',
            numpy.conjugate(States),dz,
            States)
    return d



def sort_by_state(energies, states, Nmax, Consts):
    '''Sort states by (N, M_F)_k where k is an index labelling states of
    a given N, MF in ascending energy. k is determined for the last element in 
    the array, this is usually the highest value of the varied parameter. 
    
    Args:
        energies (numpy.ndarray): eigenvalues output from numpy.linalg.eigh
        states (numpy.ndarray): eigenstates output from numpy.linalg.eigh
        Nmax: Maximum rotational quantum number in original calculations
        Consts: Molecular constants used in original calculations
        
    Output:
        energies_sorted, states_sorted: eigenenergies and eigenvalues sorted 
        by the ordering of states given in labels.
        labels: labels of the states in order
    '''
    #Find state labels
    N, I1,I2 = hamiltonian.generate_vecs(Nmax,Consts['I1'],Consts['I2'])
    
    N2 = hamiltonian.vector_dot(N,N)
    Nlabels = numpy.einsum('lik,ij,ljk->lk',numpy.conj(states),N2,states)
    Nlabels = numpy.round((-1+numpy.sqrt(1+4*1*Nlabels))/2).real
    
    F = N + I1 + I2
    Fz = F[2]
    MFlabels = numpy.round(numpy.einsum('lik,ij,ljk->lk',
                                        numpy.conj(states),Fz,states),1).real                            
                
    #Loop required to figure out k for each state                        
    labels = numpy.empty((numpy.shape(energies)[1],3))
    labels[:] = numpy.NaN
    for i in range(numpy.shape(energies)[1]):
        k = 0
        for j in range(numpy.shape(energies)[1]):
            if labels[j,0] == Nlabels[-1,i] and labels[j,1] == MFlabels[-1,i]:
                k+=1
        labels[i,:] = numpy.array([Nlabels[-1,i], MFlabels[-1,i], k])          
    
    #Now loop over energies and sort into order given by labels
    energies_sorted = numpy.zeros(numpy.shape(energies), dtype='complex')
    states_sorted = numpy.zeros(numpy.shape(states), dtype='complex')
    for i in range(numpy.shape(energies)[0]): #B field loop
        for j in  range(numpy.shape(energies)[1]): #State loop
            x = 0
            for k in range(numpy.shape(energies)[1]):
                N, MF = Nlabels[i,j], MFlabels[i,j] #N, MF for given index
                if N == labels[k,0] and MF == labels[k,1] and energies_sorted[i,k]==0 and x==0:
                    energies_sorted[i,k] = energies[i,j]
                    states_sorted[i,:,k] = states[i,:,j]
                    x += 1
                    
    return energies_sorted, states_sorted, labels


def sort_smooth(energy, states):
    ''' Sort states to remove false avoided crossings.

    This is a function to ensure that all eigenstates plotted change
    adiabatically, it does this by assuming that step to step the eigenstates
    should vary by only a small amount (i.e. that the  step size is fine) and
    arranging states to maximise the overlap one step to the next.

    Args:
        Energy (numpy.ndarray) : numpy.ndarray containing the eigenergies, as from numpy.linalg.eig
        States (numpy.ndarray): numpy.ndarray containing the states, in the same order as Energy
    Returns:
        Energy (numpy.ndarray) : numpy.ndarray containing the eigenergies, as from numpy.linalg.eig
        States (numpy.ndarray): numpy.ndarray containing the states, in the same order as Energy E[x,i] -> States[x,:,i]
    '''
    ls = numpy.arange(states.shape[2],dtype="int")
    number_iterations = len(energy[:,0])
    for i in range(1,number_iterations):
        '''
        This loop sorts the eigenstates such that they maintain some
        continuity. Each eigenstate should be chosen to maximise the overlap
        with the previous.
        '''
        #calculate the overlap of the ith and jth eigenstates
        overlaps = numpy.einsum('ij,ik->jk',
                                numpy.conjugate(states[i-1,:,:]),states[i,:,:])
        orig2 = states[i,:,:].copy()
        orig1 = energy[i,:].copy()
        #insert location of maximums into array ls
        numpy.argmax(numpy.abs(overlaps),axis=1,out=ls)
        for k in range(states.shape[2]):
            l = ls[k]
            if l!=k:
                energy[i,k] = orig1[l].copy()
                states[i,:,k] = orig2[:,l].copy()
    return energy, states
